@@ -15,8 +15,9 @@
 import traceback
 import copy
 import json
-from time import time
 import uuid
+from multiprocessing import Pool
+from time import perf_counter
 from .agent import Agent
 from .errors import DeadlineExceeded, FailedPrecondition, Internal, InvalidArgument
 from .utils import get, has, get_player, process_schema, schemas, structify
@@ -62,7 +63,7 @@ def evaluate(environment, agents=[], configuration={}, steps=[], num_episodes=1)
     return rewards
 
 
-def make(environment, configuration={}, steps=[], debug=False):
+def make(environment, configuration={}, steps=[], logs=[], debug=False, state=None):
     """
     Creates an instance of an Environment.
 
@@ -76,26 +77,37 @@ def make(environment, configuration={}, steps=[], debug=False):
         Environment: Instance of a specific environment.
     """
     if has(environment, str) and has(environments, dict, path=[environment]):
-        return Environment(**environments[environment], configuration=configuration, steps=steps, debug=debug)
+        return Environment(**environments[environment], configuration=configuration, steps=steps, logs=logs, debug=debug, state=state)
     elif callable(environment):
-        return Environment(interpreter=environment, configuration=configuration, steps=steps, debug=debug)
+        return Environment(interpreter=environment, configuration=configuration, steps=steps, logs=logs, debug=debug, state=state)
     elif has(environment, path=["interpreter"], is_callable=True):
-        return Environment(**environment, configuration=configuration, steps=steps, debug=debug)
+        return Environment(**environment, configuration=configuration, steps=steps, logs=logs, debug=debug, state=state)
     raise InvalidArgument("Unknown Environment Specification")
 
 
-class Environment:
+def act_agent(args):
+    agent, state, configuration, none_action = args
+    if state["status"] != "ACTIVE":
+        return None, {}
+    elif agent is None:
+        return none_action, {}
+    else:
+        return agent.act(state["observation"])
 
+
+class Environment:
     def __init__(
         self,
         specification={},
         configuration={},
         steps=[],
+        logs=[],
         agents={},
         interpreter=None,
         renderer=None,
         html_renderer=None,
         debug=False,
+        state=None,
     ):
         self.id = str(uuid.uuid1())
         self.debug = debug
@@ -129,23 +141,30 @@ class Environment:
             raise InvalidArgument("Default agents must be Callable.")
         self.agents = structify(agents)
 
-        if steps is None or len(steps) == 0:
-            self.reset()
-        else:
+        if steps is not None and len(steps) > 0:
             self.__set_state(steps[-1])
             self.steps = steps[0:-1] + self.steps
+        elif state is not None:
+            step = [{}] * self.specification.agents[0]
+            step[0] = state
+            self.__set_state(step)
+        else:
+            self.reset()
 
-    def step(self, actions):
+        self.logs = logs
+        self.pool = None
+
+    def step(self, actions, logs=None):
         """
         Execute the environment interpreter using the current state and a list of actions.
 
         Args:
             actions (list): Actions to pair up with the current agent states.
+            logs (list): Logs to pair up with each agent for the current step.
 
         Returns:
             list of dict: The agents states after the step.
         """
-
         if self.done:
             raise FailedPrecondition("Environment done, reset required.")
         if not actions or len(actions) != len(self.state):
@@ -179,6 +198,8 @@ class Environment:
                     s.status = "DONE"
 
         self.steps.append(self.state)
+        if logs is not None:
+            self.logs.append(logs)
 
         return self.state
 
@@ -190,7 +211,9 @@ class Environment:
             agents (list of any): List of agents to obtain actions from.
 
         Returns:
-            list of list of dict: The agent states of all steps executed.
+            tuple of:
+                list of list of dict: The agent states of all steps executed.
+                list of list of dict: The agent logs of all steps executed.
         """
         if self.state is None or len(self.steps) == 1 or self.done:
             self.reset(len(agents))
@@ -199,10 +222,10 @@ class Environment:
                 f"{len(self.state)} agents were expected, but {len(agents)} was given.")
 
         runner = self.__agent_runner(agents)
-        start = time()
-        while not self.done and time() - start < self.configuration.runTimeout:
-            self.step(runner.act())
-        runner.destroy()
+        start = perf_counter()
+        while not self.done and perf_counter() - start < self.configuration.runTimeout:
+            actions, logs = runner.act()
+            self.step(actions, logs)
         return self.steps
 
     def reset(self, num_agents=None):
@@ -258,6 +281,7 @@ class Environment:
                 "step": 0 if get(kwargs, bool, self.done, path=["autoplay"]) else (len(self.steps) - 1),
                 "controls": get(kwargs, bool, self.done, path=["controls"]),
                 "environment": self.toJSON(),
+                "logs": self.logs,
                 **kwargs,
             }
             args = [self]
@@ -265,6 +289,7 @@ class Environment:
                                      self.html_renderer(*args[:self.html_renderer.__code__.co_argcount]))
             if mode == "html":
                 return player_html
+
             from IPython.display import display, HTML
             player_html = player_html.replace('"', '&quot;')
             width = get(kwargs, int, 300, path=["width"])
@@ -274,8 +299,7 @@ class Environment:
         elif mode == "json":
             return json.dumps(self.toJSON(), sort_keys=True)
         else:
-            raise InvalidArgument(
-                "Available render modes: human, ansi, html, ipython")
+            raise InvalidArgument("Available render modes: human, ansi, html, ipython")
 
     def play(self, agents=[], **kwargs):
         """
@@ -334,26 +358,24 @@ class Environment:
 
         def advance():
             while not self.done and self.state[position].status == "INACTIVE":
-                self.step(runner.act())
+                actions, logs = runner.act()
+                self.step(actions, logs)
 
         def reset():
             nonlocal runner
             self.reset(len(agents))
-            if runner != None:
-                runner.destroy()
             runner = self.__agent_runner(agents)
             advance()
             return self.__get_shared_state(position).observation
 
         def step(action):
-            self.step(runner.act(action))
+            actions, logs = runner.act(action)
+            self.step(actions, logs)
             advance()
             agent = self.__get_shared_state(position)
             reward = agent.reward
             if len(self.steps) > 1 and reward is not None:
                 reward -= self.steps[-2][position].reward
-            if self.done:
-                runner.destroy()
             return [
                 agent.observation, reward, agent.status != "ACTIVE", agent.info
             ]
@@ -453,7 +475,6 @@ class Environment:
             }
         return structify(self.__state_schema_value)
 
-
     def __set_state(self, state=[]):
         if len(state) not in self.specification.agents:
             raise InvalidArgument(
@@ -543,48 +564,40 @@ class Environment:
         return process_schema(schemas.specification, spec)
 
     def __agent_runner(self, agents):
-        # Replace default agents with their source.
-        for i, agent in enumerate(agents):
-            if has(self.agents, path=[agent]):
-                agents[i] = self.agents[agent]
-
         # Generate the agents.
         agents = [
-            Agent(a, self.configuration, self.name, debug=self.debug)
-            if a is not None
+            Agent(agent, self)
+            if agent is not None
             else None
-            for a in agents
+            for agent in agents
         ]
 
-        # Have the agents had a chance to initialize (first non-empty act).
-        initialized = [False] * len(agents)
-
         def act(none_action=None):
             if len(agents) != len(self.state):
                 raise InvalidArgument(
                     "Number of agents must match the state length")
 
-            actions = [0] * len(agents)
-            for i, agent in enumerate(agents):
-                if self.state[i]["status"] != "ACTIVE":
-                    actions[i] = None
-                elif agent == None:
-                    actions[i] = none_action
-                else:
-                    timeout = self.configuration.actTimeout
-                    if not initialized[i]:
-                        initialized[i] = True
-                        timeout += self.configuration.agentTimeout
-                    state = self.__get_shared_state(i)
-                    actions[i] = agent.act(state, timeout)
-            return actions
+            act_args = [
+                (
+                    agent,
+                    self.__get_shared_state(i),
+                    self.configuration,
+                    none_action,
+                )
+                for i, agent in enumerate(agents)
+            ]
 
-        def destroy():
-            for a in agents:
-                if a != None:
-                    a.destroy()
+            if all((agent is None or agent.is_parallelizable) for agent in agents):
+                if self.pool is None:
+                    self.pool = Pool(processes=len(agents))
+                results = self.pool.map(act_agent, act_args)
+            else:
+                results = list(map(act_agent, act_args))
 
-        return structify({"act": act, "destroy": destroy})
+            actions, logs = zip(*results)
+            return actions, logs
+
+        return structify({"act": act})
 
     def __get_shared_state(self, position):
         if position == 0:
