@@ -230,3 +230,10 @@ def test_deadline():
         }
     ]
 
+
+def test_render():
+    before_each(configuration={"team_1": 1, "team_2": 1, "scenario_name": "tests.penalty", "save_video": True})
+    env.step([[0],[0]])
+    env.render(mode="human", width=800, height=600)
+    env.step([[0],[0]])
+    env.render(mode="human", width=800, height=600)
