@@ -1,7 +1,23 @@
+# Copyright 2020 Kaggle Inc
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 from copy import deepcopy
 from enum import Enum, auto
+from functools import wraps
 from typing import *
 import operator
+import sys
 
 
 # region Helper Classes and Methods
@@ -164,11 +180,6 @@ class Configuration(ReadOnlyDict[str, any]):
         """Total number of steps/turns in the run."""
         return self["episodeSteps"]
 
-    @property
-    def agent_exec(self) -> str:
-        """How the agent is executed alongside the running envionment ('LOCAL' or separate 'PROCESS')."""
-        return self["agentExec"]
-
     @property
     def agent_timeout(self) -> float:
         """Maximum runtime (seconds) to initialize an agent."""
@@ -557,7 +568,7 @@ class Board:
                 raw_action = player_actions.get(ship_id)
                 action = (
                     ShipAction[raw_action]
-                    if raw_action is not None
+                    if raw_action in ShipAction.__members__
                     else None
                 )
                 self._add_ship(Ship(ship_id, ship_position, ship_halite, player_id, self, action))
@@ -567,7 +578,7 @@ class Board:
                 raw_action = player_actions.get(shipyard_id)
                 action = (
                     ShipyardAction[raw_action]
-                    if raw_action is not None
+                    if raw_action in ShipyardAction.__members__
                     else None
                 )
                 self._add_shipyard(Shipyard(shipyard_id, shipyard_position, player_id, self, action))
@@ -831,8 +842,12 @@ def board_agent(agent: Callable[[Board], None]):
     def my_agent(board: Board) -> None:
         ...
     """
+    @wraps(agent)
     def agent_wrapper(obs, config) -> Dict[str, str]:
         board = Board(obs, config)
         agent(board)
         return board.current_player.next_actions
+
+    if agent.__module__ is not None and agent.__module__ in sys.modules:
+        setattr(sys.modules[agent.__module__], agent.__name__, agent_wrapper)
     return agent_wrapper
