@@ -15,6 +15,7 @@
 import argparse
 import json
 import traceback
+from typing import *
 from . import errors, utils
 from .agent import Agent
 from .core import environments, evaluate, make
@@ -23,7 +24,7 @@ from logging.config import dictConfig
 parser = argparse.ArgumentParser(description="Kaggle Simulations")
 parser.add_argument(
     "action",
-    choices=["list", "evaluate", "run", "step", "load", "act", "http-server"],
+    choices=["list", "evaluate", "run", "step", "load", "act", "dispose", "http-server"],
     help="List environments. Evaluate many episodes. Run a single episode. Step the environment. Load the environment. Start http server.",
 )
 parser.add_argument("--environment", type=str,
@@ -42,6 +43,11 @@ parser.add_argument(
     type=json.loads,
     help="Environment starting states (default=[resetState]).",
 )
+parser.add_argument(
+    "--logs",
+    type=json.loads,
+    help="Environment starting logs (default=[]).",
+)
 parser.add_argument(
     "--state",
     type=json.loads,
@@ -53,10 +59,12 @@ parser.add_argument(
 parser.add_argument(
     "--render",
     type=json.loads,
-    help="Response from run, step, or load. Calls environment render. (default={mode='json'})",
+    help="Response from run, step, or load. Calls environment render (default={mode='json'}).",
 )
 parser.add_argument(
-    "--middleware", type=str, help="Path to request middleware for use with the http-server."
+    "--display",
+    type=str,
+    help="Shortcut to the --render {mode=''} argument (default json).",
 )
 parser.add_argument(
     "--port", type=int, help="http-server Port (default=8000)."
@@ -65,12 +73,22 @@ parser.add_argument(
     "--host", type=str, help="http-server Host (default=127.0.0.1)."
 )
 parser.add_argument(
-    "--out", type=str, help="Output file to write the results of the episode."
+    "--in", type=str, help="Episode replay file to load. Only works when the action is load."
+)
+parser.add_argument(
+    "--out", type=str, help="Output file to write the results of the episode. Does nothing in http-server mode."
+)
+parser.add_argument(
+    "--log", type=str, help="Agent log file to write the std out, resource, and step timing for each agent. Also used to load logs from a file with the load action."
 )
 
 
 def render(args, env):
-    mode = utils.get(args.render, str, "json", path=["mode"])
+    mode = \
+        args.display \
+        if args.display is not None \
+        else utils.get(args.render, str, "json", path=["mode"])
+
     if mode == "human" or mode == "ansi":
         args.render["mode"] = "ansi"
     elif mode == "ipython" or mode == "html":
@@ -102,52 +120,78 @@ def action_act(args):
         return {"error": "One agent must be provided."}
     raw = args.agents[0]
 
-    # Pass empty steps in here because we just need the configuration from the environment
-    env = make(args.environment, args.configuration, [], args.debug)
-    config = env.configuration
-    timeout = config.actTimeout
+    env = make(args.environment, args.configuration, state=args.state, debug=args.debug)
 
-    if cached_agent is None or cached_agent.id != raw:
-        if cached_agent is not None:
-            cached_agent.destroy()
-        identifier = raw
-        if utils.has(env.agents, path=[identifier]):
-            raw = env.agents[identifier]
-        cached_agent = Agent(raw, config, args.environment, identifier, args.debug)
-        timeout = config.agentTimeout
-    state = {
-        "observation": utils.get(args.state, dict, {}, ["observation"]),
-        "reward": args.get("reward", None),
-        "info": utils.get(args.state, dict, {}, ["info"])
-    }
-    action = cached_agent.act(state, timeout)
+    is_first_run = cached_agent is None or cached_agent.raw != raw
+    if is_first_run:
+        cached_agent = Agent(raw, env)
+    observation = utils.get(args.state, dict, {}, ["observation"])
+    action, log = cached_agent.act(observation)
     if isinstance(action, errors.DeadlineExceeded):
         action = "DeadlineExceeded"
     elif isinstance(action, BaseException):
         action = "BaseException::" + str(action)
 
+    if args.log_path is not None:
+        with open(args.log_path, mode="a") as log_file:
+            if not is_first_run:
+                log_file.write(",\n ")
+            json.dump([log], log_file)
+
     return {"action": action}
 
 
 def action_step(args):
-    env = make(args.environment, args.configuration, args.steps, args.debug)
+    env = make(args.environment, args.configuration, args.steps, args.logs, args.debug)
     runner = env.__agent_runner(args.agents)
     env.step(runner.act())
-    runner.destroy()
+    if args.log_path is not None:
+        with open(args.log_path, mode="a") as log_file:
+            json.dump(env.logs[-1], log_file)
+            log_file.write(",")
     return render(args, env)
 
 
 def action_run(args):
-    env = make(args.environment, args.configuration, args.steps, args.debug)
+    env = make(args.environment, args.configuration, args.steps, args.logs, args.debug)
     env.run(args.agents)
+    if args.log_path is not None:
+        with open(args.log_path, mode="w") as log_file:
+            json.dump(env.logs, log_file)
     return render(args, env)
 
 
 def action_load(args):
-    env = make(args.environment, args.configuration, args.steps, args.debug)
+    if args.log_path is not None:
+        with open(args.log_path, mode="r") as log_file:
+            args.logs = json.load(log_file)
+
+    if args.in_path is not None:
+        with open(args.in_path, mode="r") as replay_file:
+            json_args = json.load(replay_file)
+        env = make(json_args["name"], json_args["configuration"], json_args["steps"], args.logs, args.debug)
+    else:
+        env = make(args.environment, args.configuration, args.steps, args.logs, args.debug)
     return render(args, env)
 
 
+disposed = True
+
+
+# This method is only called at the end of an episode to write the final array brace in the logs file and dispose the cached agent to force reinitialization
+def action_dispose(args):
+    global cached_agent, disposed
+    if disposed:
+        return "Already disposed"
+
+    cached_agent = None
+    if args.log_path is not None:
+        with open(args.log_path, mode="a") as log_file:
+            log_file.write("]")
+    disposed = True
+    return "Successfully disposed"
+
+
 def parse_args(args):
     return utils.structify(
         {
@@ -158,11 +202,15 @@ def parse_args(args):
             "episodes": utils.get(args, int, 1, ["episodes"]),
             "state": utils.get(args, dict, {}, ["state"]),
             "steps": utils.get(args, list, [], ["steps"]),
+            "logs": utils.get(args, list, [], ["logs"]),
             "render": utils.get(args, dict, {"mode": "json"}, ["render"]),
+            "display": utils.get(args, str, None, ["display"]),
             "debug": utils.get(args, bool, False, ["debug"]),
             "host": utils.get(args, str, "127.0.0.1", ["host"]),
             "port": utils.get(args, int, 8000, ["port"]),
-            "out": utils.get(args, str, None, ["out"]),
+            "in_path": utils.get(args, str, None, ["in"]),
+            "out_path": utils.get(args, str, None, ["out"]),
+            "log_path": utils.get(args, str, None, ["log"]),
         }
     )
 
@@ -171,31 +219,40 @@ def action_handler(args):
     try:
         if args.action == "list":
             return action_list(args)
-        elif args.action == "http-server":
+        if args.action == "http-server":
             return {"error": "Already running a http server."}
-        elif args.action == "act":
+        if args.action == "act":
             return action_act(args)
+        if args.action == "dispose":
+            return action_dispose(args)
+        if args.action == "load":
+            return action_load(args)
 
-        if args.environment == None:
+        if args.environment is None:
             return {"error": "Environment required."}
 
         if args.action == "evaluate":
             return action_evaluate(args)
-        elif args.action == "step":
+        if args.action == "step":
             return action_step(args)
-        elif args.action == "run":
+        if args.action == "run":
             return action_run(args)
-        elif args.action == "load":
-            return action_load(args)
-        else:
-            return {"error": "Unknown Action"}
+
+        return {"error": "Unknown Action"}
     except Exception as e:
         return {"error": str(e), "trace": traceback.format_exc()}
 
 
+log_path: Optional[str] = None
+
+
 def action_http(args):
     from flask import Flask, request
 
+    if args.log_path is not None:
+        global log_path
+        log_path = args.log_path
+
     # Setup logging to console for Flask
     dictConfig({
         'version': 1,
@@ -213,25 +270,12 @@ def action_http(args):
         }
     })
 
-    middleware = {"request": None, "response": None}
-    if args.middleware != None:
-        try:
-            raw = utils.read_file(args.middleware)
-            local = utils.get_exec(raw)
-            middleware["request"] = utils.get(
-                local, path=["request"], is_callable=True)
-            middleware["response"] = utils.get(
-                local, path=["response"], is_callable=True)
-        except Exception as e:
-            return {"error": str(e), "trace": traceback.format_exc()}
-
     app = Flask(__name__, static_url_path="", static_folder="")
-    app.route("/", methods=["GET", "POST"]
-              )(lambda: http_request(request, middleware))
+    app.route("/", methods=["GET", "POST"])(lambda: http_request(request))
     app.run(args.host, args.port, debug=True)
 
 
-def http_request(request, middleware):
+def http_request(request):
     # Set CORS headers for the preflight request
     if request.method == "OPTIONS":
         # Allows GET requests from any origin with the Content-Type
@@ -246,7 +290,6 @@ def http_request(request, middleware):
         return "", 204, headers
 
     headers = {"Access-Control-Allow-Origin": "*"}
-
     params = request.args.to_dict()
     for key in list(params.keys()):
         if key.endswith("[]"):
@@ -257,28 +300,36 @@ def http_request(request, middleware):
             del params[key]
 
     body = request.get_json(silent=True, force=True) or {}
+    args = {**params, **body}
+    if "render" in args:
+        # Manually deserialize render argument
+        # We should eventually refactor this to use the same deserializer as the cmd line arg parser
+        args["render"] = json.loads(args["render"])
+    args = parse_args(args)
+    if args.log_path is None:
+        args.log_path = log_path
 
-    req = parse_args({**params, **body})
-    if middleware["request"] is not None:
-        req = middleware["request"](req)
-
-    resp = action_handler(req)
-    if middleware["response"] is not None:
-        resp = middleware["response"](req, resp)
+    global disposed
+    # Write the opening array brace for the logs file if there is a logs file.
+    if disposed and args["action"] != "dispose" and args.log_path is not None:
+        with open(args.log_path, mode="w") as log_file:
+            log_file.write("[")
+        disposed = False
 
+    resp = action_handler(args)
     return resp, 200, headers
 
 
 def main():
-    args = parser.parse_args()
+    args = parse_args(vars(parser.parse_args()))
     if args.action == "http-server":
         action_http(args)
     else:
-        result = action_handler(parse_args(vars(args)))
-        if args.out is None:
+        result = action_handler(args)
+        if args.out_path is None:
             print(result)
         else:
-            with open(args.out, mode="w") as out_file:
+            with open(args.out_path, mode="w") as out_file:
                 out_file.write(str(result))
 
         return 0
